import * as ts from 'typescript';
export interface INodeTypeScriptServiceOptions {
    /**
     * TypeScript configuration file name.
     *
     * @default 'tsconfig.json'
     */
    tsconfigFileName?: string;
    /**
     * Compiler options to use when transpiling files using tsconfig.
     * Overrides options defined in `tsconfig.json` files.
     *
     * @default {}
     */
    overrideOptions?: ts.CompilerOptions;
    /**
     * Default compiler options to use when no tsconfig is found.
     *
     * @default {}
     */
    noConfigOptions?: ts.CompilerOptions;
    /**
     * Transformers to apply during transpilation
     */
    customTransformers?: ts.CustomTransformers;
}
export interface ITranspilationOutput {
    /** Absolute file path to the input typescript file */
    filePath: string;
    /** transpiled JavaScript code */
    outputText: string;
    /** optional, separate source-maps (stringified JSON) */
    sourceMapText?: string;
    /** Transpilation process diagnostics  */
    diagnostics?: ts.Diagnostic[];
}
/**
 * On-demand TypeScript tranpilation service
 * Options can be provided during construction
 *
 * @example
 * new NodeTypeScriptService()
 * new NodeTypeScriptService({tsconfigFileName: 'tsconfig.test.json'})
 */
export declare class NodeTypeScriptService {
    options: Required<INodeTypeScriptServiceOptions>;
    runningServices: Map<string, ts.LanguageService>;
    private sharedDocumentRegistry;
    private directoryToTsConfig;
    constructor(options?: INodeTypeScriptServiceOptions);
    /**
     * Load a `tsconfig.json` file by providing a path to it.
     * A language service will be registered for th
     * @param configFilePath absolute path to the configuration file (usually `tsconfig.json`)
     */
    loadConfigFile(configFilePath: string): {
        diagnostics: ts.Diagnostic[];
        fileNames: string[];
        options: ts.CompilerOptions;
        languageService: ts.LanguageService;
    };
    /**
     * Transpile a TypeScript file on the native file system
     *
     * @param filePath absolute path of the source file to transpile
     */
    transpileFile(filePath: string): ITranspilationOutput;
    private transpileUsingLanguageService;
    private transpileUsingDefaultOptions;
    /**
     * Find the closest `tsconfig.json` file to the provided baseDirectory
     *
     * @param baseDirectory the directory to start looking from
     */
    private getTsConfigPath;
    private createLanguageServiceHost;
    private createErrorDiagnostic;
}
//# sourceMappingURL=node-ts-service.d.ts.map