"use strict";
var __importStar = (this && this.__importStar) || function (mod) {
    if (mod && mod.__esModule) return mod;
    var result = {};
    if (mod != null) for (var k in mod) if (Object.hasOwnProperty.call(mod, k)) result[k] = mod[k];
    result["default"] = mod;
    return result;
};
Object.defineProperty(exports, "__esModule", { value: true });
const fs_1 = require("fs");
const path_1 = require("path");
const ts = __importStar(require("typescript"));
const defaultOptions = {
    tsconfigFileName: 'tsconfig.json',
    overrideOptions: {},
    noConfigOptions: {},
    customTransformers: {}
};
/**
 * On-demand TypeScript tranpilation service
 * Options can be provided during construction
 *
 * @example
 * new NodeTypeScriptService()
 * new NodeTypeScriptService({tsconfigFileName: 'tsconfig.test.json'})
 */
class NodeTypeScriptService {
    constructor(options) {
        // a map holding `tsconfig path` to a `language service` instance
        this.runningServices = new Map();
        // we might create more than a single language service per service, so we share documents between them
        this.sharedDocumentRegistry = ts.createDocumentRegistry(ts.sys.useCaseSensitiveFileNames, ts.sys.getCurrentDirectory());
        // cache of `directory path` to `tsconfig lookup result`, to save disk operations
        this.directoryToTsConfig = new Map();
        this.options = Object.assign({}, defaultOptions, options);
    }
    /**
     * Load a `tsconfig.json` file by providing a path to it.
     * A language service will be registered for th
     * @param configFilePath absolute path to the configuration file (usually `tsconfig.json`)
     */
    loadConfigFile(configFilePath) {
        const jsonSourceFile = ts.readJsonConfigFile(configFilePath, ts.sys.readFile);
        const configDirectoryPath = path_1.dirname(configFilePath);
        const { fileNames: originalFileNames, options, errors: diagnostics } = ts.parseJsonSourceFileConfigFileContent(jsonSourceFile, ts.sys, configDirectoryPath);
        const fileNames = originalFileNames.map(path_1.normalize);
        // create the host
        const languageServiceHost = this.createLanguageServiceHost(fileNames, Object.assign({}, options, this.options.overrideOptions));
        // create the language service using the host
        const languageService = ts.createLanguageService(languageServiceHost, this.sharedDocumentRegistry);
        // register it in our running services
        this.runningServices.set(configFilePath, languageService);
        return {
            diagnostics,
            fileNames,
            options,
            languageService
        };
    }
    /**
     * Transpile a TypeScript file on the native file system
     *
     * @param filePath absolute path of the source file to transpile
     */
    transpileFile(filePath) {
        for (const existingLanguageService of this.runningServices.values()) {
            const program = existingLanguageService.getProgram();
            if (program && program.getRootFileNames().includes(filePath)) {
                return this.transpileUsingLanguageService(filePath, existingLanguageService);
            }
        }
        const fileDirectoryPath = path_1.dirname(filePath);
        const tsConfigPath = this.getTsConfigPath(fileDirectoryPath);
        if (!tsConfigPath) {
            return this.transpileUsingDefaultOptions(filePath);
        }
        const { fileNames, diagnostics, languageService } = this.loadConfigFile(tsConfigPath);
        if (diagnostics.length) {
            return {
                diagnostics,
                filePath,
                outputText: ''
            };
        }
        // verify the new service includes our file
        if (fileNames.includes(filePath)) {
            return this.transpileUsingLanguageService(filePath, languageService);
        }
        return this.transpileUsingDefaultOptions(filePath);
    }
    transpileUsingLanguageService(filePath, languageService) {
        const { outputFiles, emitSkipped } = languageService.getEmitOutput(filePath);
        if (emitSkipped) {
            return {
                filePath,
                diagnostics: [
                    this.createErrorDiagnostic('Emit was skipped')
                ],
                outputText: ''
            };
        }
        const jsOutputFile = outputFiles.filter(outputFile => outputFile.name.endsWith('.js')).shift();
        if (!jsOutputFile) {
            return {
                filePath,
                diagnostics: [
                    this.createErrorDiagnostic('No js output file was found')
                ],
                outputText: ''
            };
        }
        const sourceMapOutputFile = outputFiles.filter(outputFile => outputFile.name.endsWith('.js.map')).shift();
        const sourceMapText = sourceMapOutputFile && sourceMapOutputFile.text;
        const syntacticDiagnostics = languageService.getSyntacticDiagnostics(filePath);
        if (syntacticDiagnostics.length) {
            return {
                diagnostics: syntacticDiagnostics,
                filePath,
                outputText: jsOutputFile.text,
                sourceMapText
            };
        }
        const semanticDiagnostics = languageService.getSemanticDiagnostics(filePath);
        if (semanticDiagnostics.length) {
            return {
                diagnostics: semanticDiagnostics,
                filePath,
                outputText: jsOutputFile.text,
                sourceMapText
            };
        }
        return {
            filePath,
            outputText: jsOutputFile.text,
            sourceMapText
        };
    }
    transpileUsingDefaultOptions(filePath) {
        const tsCode = ts.sys.readFile(filePath);
        if (!tsCode) {
            return {
                filePath,
                diagnostics: [
                    this.createErrorDiagnostic(`Unable to read ${filePath}`)
                ],
                outputText: ''
            };
        }
        const { outputText, diagnostics, sourceMapText } = ts.transpileModule(tsCode, {
            fileName: filePath,
            compilerOptions: this.options.noConfigOptions,
            transformers: this.options.customTransformers
        });
        return {
            filePath,
            outputText,
            sourceMapText,
            diagnostics
        };
    }
    /**
     * Find the closest `tsconfig.json` file to the provided baseDirectory
     *
     * @param baseDirectory the directory to start looking from
     */
    getTsConfigPath(baseDirectory) {
        if (this.directoryToTsConfig.has(baseDirectory)) {
            return this.directoryToTsConfig.get(baseDirectory);
        }
        const tsConfigPath = ts.findConfigFile(baseDirectory, ts.sys.fileExists, this.options.tsconfigFileName);
        this.directoryToTsConfig.set(baseDirectory, tsConfigPath);
        return tsConfigPath;
    }
    createLanguageServiceHost(fileNames, compilerOptions) {
        return {
            getCompilationSettings: () => compilerOptions,
            getNewLine: () => {
                switch (compilerOptions.newLine) {
                    case ts.NewLineKind.CarriageReturnLineFeed:
                        return '\r\n';
                    case ts.NewLineKind.LineFeed:
                        return '\n';
                    default:
                        return ts.sys.newLine;
                }
            },
            getScriptFileNames: () => fileNames,
            getScriptVersion: filePath => {
                try {
                    return fs_1.statSync(filePath).mtime.toString();
                }
                catch (_a) {
                    return `${Date.now()}`;
                }
            },
            getScriptSnapshot: filePath => ts.ScriptSnapshot.fromString(ts.sys.readFile(filePath) || ''),
            getCurrentDirectory: ts.sys.getCurrentDirectory,
            getDefaultLibFileName: ts.getDefaultLibFilePath,
            useCaseSensitiveFileNames: () => ts.sys.useCaseSensitiveFileNames,
            readDirectory: ts.sys.readDirectory,
            readFile: ts.sys.readFile,
            fileExists: ts.sys.fileExists,
            directoryExists: ts.sys.directoryExists,
            getDirectories: ts.sys.getDirectories,
            realpath: ts.sys.realpath,
            getCustomTransformers: () => this.options.customTransformers
        };
    }
    createErrorDiagnostic(messageText) {
        return {
            messageText,
            category: ts.DiagnosticCategory.Error,
            code: ts.DiagnosticCategory.Error,
            file: undefined,
            start: 0,
            length: undefined
        };
    }
}
exports.NodeTypeScriptService = NodeTypeScriptService;
//# sourceMappingURL=node-ts-service.js.map